/*
Copyright (c) 2009, INRIA, Université de Nancy 2 and Universidade
Federal do Rio Grande do Norte.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
   * Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
   * Neither the name of the Université de Nancy 2 or the Universidade Federal
     do Rio Grande do Norte nor the names of its contributors may be used
     to endorse or promote products derived from this software without
     specific prior written permission.

THIS SOFTWARE IS PROVIDED BY INRIA, Université de Nancy 2 and
Universidade Federal do Rio Grande do Norte ''AS IS'' AND ANY EXPRESS
OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL INRIA, Université de Nancy 2 and
Universidade Federal do Rio Grande do Norte BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

*/

/**
   \file tptp-logic.c

   --------------------------------------------------------------
*/

#include <stdlib.h>
#include <string.h>

#include "general.h"
#include "table.h"

#include "DAG.h"

/*--------------------------------------------------------------*/

Tsort  tptp_sort;

/** \brief A sort for E terms when client logic is multi-sorted */
static Tsort    my_sort;

int    TPTP_SYMB_SORT      = 1;
int    TPTP_SYMB_SKOLEM    = 2;
int    TPTP_SYMB_PREDICATE = 3;

/** \brief The name for the sort of TPTP terms */
static char     tptp_sort_str[] = "veriT__TPTP";
/** \brief Table of pre-computed signatures for TPTP function symbols */
static Ttable   table_tptp_fun_sig;
/** \brief Table of pre-computed signatures for TPTP predicate symbols */
static Ttable   table_tptp_pred_sig;
/** \brief Table of skolem symbols generated by the TPTP prover */
static Ttable   table_skolem;
/** \brief Table of epred symbols generated by the TPTP prover */
static Ttable   table_epred;

/*--------------------------------------------------------------*/

void
tptp_logic_init(void)
{
  my_sort = DAG_sort_new(tptp_sort_str, 0, NULL);
  DAG_sort_set_predefined(my_sort);
  table_tptp_fun_sig = table_new(4, 1);
  table_tptp_pred_sig = table_new(4, 1);
  table_skolem = table_new(2, 1);
  table_epred = table_new(2, 1);
}

/*--------------------------------------------------------------*/

void
tptp_logic_quit(void)
{
  int i;
  for (i = 0; i < table_length(table_skolem); ++i)
    {
      Tstack_symb symbols = (Tstack_symb) table_get(table_skolem, i);
      stack_free(symbols);
    }
  for (i = 0; i < table_length(table_epred); ++i)
    {
      Ttable table2 = (Ttable) table_get(table_skolem, i);
      table_free(&table2);
    }
  table_free(&table_tptp_fun_sig);
  table_free(&table_tptp_pred_sig);
  table_free(&table_skolem);
  table_free(&table_epred);
}

/*--------------------------------------------------------------*/

void
tptp_logic_set_sort (Tsort sort)
{
  int i;
  if ((sort == DAG_SORT_NULL && tptp_sort == my_sort) || sort == tptp_sort) 
    return;
  if (sort == DAG_SORT_NULL)
    tptp_sort = my_sort;
  else
    tptp_sort = sort;
  for (i = 0; i < table_length(table_skolem); ++i)
    {
      Tstack_symb symbols = (Tstack_symb) table_get(table_skolem, i);
      stack_free(symbols);
    }
  for (i = 0; i < table_length(table_epred); ++i)
    {
      Ttable table2 = (Ttable) table_get(table_skolem, i);
      table_free(&table2);
    }
  table_erase(table_tptp_fun_sig);
  table_erase(table_tptp_pred_sig);
  table_erase(table_skolem);
  table_erase(table_epred);
  /* sort for functions of arity - 1*/
  table_push(table_tptp_fun_sig, DAG_ptr_of_sort(DAG_SORT_NULL));
  /* for constants */
  table_push(table_tptp_fun_sig, DAG_ptr_of_sort(tptp_sort)); 
  table_push(table_tptp_pred_sig, DAG_ptr_of_sort(SORT_BOOLEAN));
}

/*--------------------------------------------------------------*/

/**
   \brief Returns sort of TPTP function application with arity
   \param n the arity
   \note  The results are cached in table_tptp_fun_sig.
*/

static Tsort
tptp_fun_sig(int n)
{
  /* layout of table_tptp_fun_sig:
     table_tptp_fun_sig[0] -> arity -1 -> length == 1
     table_tptp_fun_sig[1] -> arity 0 -> length == 2
     table_tptp_fun_sig[2] -> arity 1 -> length == 3
     table_tptp_fun_sig[n+1] -> arity n -> length == n+2 
  */
  assert(table_length(table_tptp_fun_sig) > 1);
  if (n == -1)
    {
      Tsort sort = DAG_sort_of_ptr(table_get(table_tptp_fun_sig, 0));
      if (sort == DAG_SORT_NULL)
	{
	  Tsort *Psort;
	  MY_MALLOC(Psort, 2*sizeof(Tsort));
	  Psort[0] = Psort[1] = tptp_sort;
	  sort = DAG_sort_new (NULL, -1, Psort);
	  table_set(table_tptp_fun_sig, 0, DAG_ptr_of_sort(sort));
	}
      return sort;
    }
  if (table_length(table_tptp_fun_sig) < n + 2)
    {
      int i, j;
      for (i = table_length(table_tptp_fun_sig); i <= n+1; ++i)
	{
	  Tsort *sub;
	  MY_MALLOC(sub, i*sizeof(Tsort));
	  for (j = 0; j <= i-1; ++j)
	    sub[j] = tptp_sort;
	  table_push(table_tptp_fun_sig, 
		     DAG_ptr_of_sort(DAG_sort_new(NULL, i, sub)));
	}
    }
  assert(table_length(table_tptp_fun_sig) >= n+2);
  return DAG_sort_of_ptr(table_get(table_tptp_fun_sig, n+1));
}

/*--------------------------------------------------------------*/

/**
   \brief Returns sort of TPTP predicate application with arity
   \param n the arity
   \note  The results are cached in table_tptp_sig.
*/
static Tsort
tptp_pred_sig(int n)
{
  assert (n >= 0);
  /* table_tptp_pred_sig[0] -> arity 0 -> length == 1 */
  /* table_tptp_pred_sig[1] -> arity 1 -> length == 2 */
  /* table_tptp_pred_sig[n] -> arity n -> length == n+1 */
  assert(table_length(table_tptp_pred_sig) > 0);
  if (table_length(table_tptp_pred_sig) < n + 1)
    {
      int i, j;
      for (i = table_length(table_tptp_pred_sig); i <= n; ++i)
	{
	  Tsort *sub;
	  MY_MALLOC(sub, (i+1) *sizeof(Tsort));
	  for (j = 0; j < i; ++j)
	    sub[j] = tptp_sort;
	  sub[i] = SORT_BOOLEAN;
	  table_push(table_tptp_pred_sig, 
		     DAG_ptr_of_sort(DAG_sort_new(NULL, i+1, sub)));
	}
    }
  assert(table_length(table_tptp_pred_sig) >= n+1);
  return DAG_sort_of_ptr(table_get(table_tptp_pred_sig, n));
}

/*--------------------------------------------------------------*/

Tsymb tptp_logic_skolem(int arity, int count)
{
  static char name_prefix [] = "veriT__e_sk_";
  Tstack_symb symbols;
  if (table_length(table_skolem) < arity + 1)
    {
      int i;
      for (i = table_length(table_skolem); i <= arity; ++i)
	{
	  stack_INIT(symbols);
	  table_push(table_skolem, symbols);
	}
    }
  symbols = (Tstack_symb) table_get(table_skolem, arity);
  if (symbols->size < count + 1)
    {
      int i, begin;
      begin = symbols->size;
      for (i = begin; i <= count; ++i)
	{
	  char * name;
	  Tsymb symb;
	  MY_MALLOC(name, strlen(name_prefix)+ul_str_size(arity)+
		    ul_str_size(count)+2);
	  sprintf(name, "%s%i_%i", name_prefix, arity, i);
	  symb = DAG_symb_new(name, SYMB_SKOLEM, tptp_fun_sig(arity));
	  free(name);
	  stack_push(symbols, symb);
	}
      table_set(table_skolem, arity, symbols);
    }
  return symbols->data[count];
}

/*--------------------------------------------------------------*/

Tsymb tptp_logic_epred(int count, int arity)
{
  static char name_prefix [] = "veriT__e_pred_";
  Tstack_symb symbols;
  if (table_length(table_epred) < count + 1)
    {
      int i;
      for (i = table_length(table_epred); i <= count; ++i)
	{
	  stack_INIT(symbols);
	  table_push(table_epred, symbols);
	}
    }
  symbols = (Tstack_symb) table_get(table_epred, count);
  if (symbols->size < arity + 1)
    {
      int i;
      for (i = symbols->size; i <= arity; ++i)
	{
	  char * name;
	  Tsymb symb;
	  MY_MALLOC(name, strlen(name_prefix)+ul_str_size(count)+
		    ul_str_size(arity)+2);
	  sprintf(name, "%s%i_%i", name_prefix, count, arity);
	  symb = DAG_symb_new(name, SYMB_PREDICATE, tptp_pred_sig(arity));
	  free(name);
	  stack_push(symbols, symb);
	}
    }
  return symbols->data[arity];
}

/*--------------------------------------------------------------*/

static void 
tptp_mark_symbol_aux(Ttable table, int value)
{
  unsigned i, j;
  for (i = 0; i < table_length(table); ++i)
    {
      Tstack_symb symbols = (Tstack_symb) table_get(table, i);
      for (j = 0; j < symbols->size; j++)
	DAG_symb_set_misc(symbols->data[j], value);
    }
}

/*--------------------------------------------------------------*/

void 
tptp_mark_symbols(void)
{
  tptp_mark_symbol_aux(table_skolem, TPTP_SYMB_SKOLEM);
  tptp_mark_symbol_aux(table_epred, TPTP_SYMB_PREDICATE);
}


/*--------------------------------------------------------------*/

void 
tptp_unmark_symbols(void)
{
  tptp_mark_symbol_aux(table_skolem, 0);
  tptp_mark_symbol_aux(table_epred, 0);
}

/*--------------------------------------------------------------*/

Tsymb
tptp_logic_function(char * name, int arity)
{
  if (name == NULL)
    return DAG_symb_new("veriT__fn", SYMB_FUNCTION, tptp_fun_sig(arity));
  else
    return DAG_symb_new(name, SYMB_FUNCTION, tptp_fun_sig(arity));
}

/*--------------------------------------------------------------*/

Tsymb
tptp_logic_predicate(char * name, int arity)
{
  return DAG_symb_new(name, SYMB_PREDICATE, tptp_pred_sig(arity));
}

/*--------------------------------------------------------------*/

Tsymb 
tptp_logic_const(void)
{
  return DAG_symb_const(tptp_sort);
}
